
"""
Helper para o profiling de programas.
"""

import time
import inspect

class ProfilingPerformance():
    """
    Classe mix-in para profiling da performance do código.

    Grava o tempo gasto entre o início e o fim da contagem de acordo com a
    configuração feita no método configura_profiling.
    """

    # <atributos de classe>

    # guarda a contagem de quantos profilings estão sendo realizados, para
    # fechar o arquivo de profiling quando o última profiling for destruído
    __qtd_profilings = 0

    # planilha onde serão gravados os dados de profiling coletados
    __planilha = None

    def __del__(self):
        # se for a última instância de profiling sendo desruída, fecha a
        # planilha de saída
        if self._profiling_on:
            ProfilingPerformance.__qtd_profilings -= 1

            if ProfilingPerformance.__qtd_profilings == 0:
                ProfilingPerformance.__planilha.close()

    def habilita_profiling_performance(self, path_csv=""):
        """
        Habilita o registro da performance de execução. O resultado do profiling
        será gravado na planilha 'performance.csv'.

        Args:
            path_csv : caminho onde deverá ser salva a planilha. Se não
                       informado, será salva no diretório atual
            
        """

        if ProfilingPerformance.__qtd_profilings == 0:
            # TODO: acrescentar path (que é um saco!)
            try:
                ProfilingPerformance.__planilha = open("performance.csv", "a")
                # escreve uma linha em branco para separar os registros entre
                # uma sessão de profiling e outra
                ProfilingPerformance.__planilha.write(
                        "Funcao;" 
                        "Objeto;"
                        "Tempo em segundos\n")
            except Exception as err:
                raise SystemExit(err)

        ProfilingPerformance.__qtd_profilings += 1
        self._profiling_on = True

    def _inicia_timer(self):
        """
        Inicia o contador de tempo.
        """

        self.__inicio = time.perf_counter()

    def _finaliza_timer(self):
        """
        Calcula o tempo despendido.
        """
        duracao = time.perf_counter() - self.__inicio
        self.__grava_resultado(duracao)

    def __grava_resultado(self, duracao):
        #  pega o nome do método para o qual está sendo realizado o profiling
        funcao = inspect.stack()[2].function

        ProfilingPerformance.__planilha.write(
                funcao + ";" + \
                str(self) + ";" +  \
                ("%10.10f" % duracao).replace(".", ",") + "\n")




import blipy.planilha as planilha


## Teste 01 __________________________________________
def teste_01(param_io, param_sheet_name):
    print('Teste 01')
    print('Execução básica - print de registros da tabela.')
    print()

    p = planilha.Planilha()
    p.carrega_dados(io=param_io, sheet_name=param_sheet_name)

    for i in range(25):
        r = p.le_prox_registro()
        print(r)


    print()
    print('Fim teste 01')


# Teste 02 __________________________________________
def teste_02(param_io, param_sheet_name):
    print('Teste 02')
    print('Duas execuções conseguintes.\n'
          'Entre a primeira e a segunda é executada a função recarrega_dados()')
    print()

    p = planilha.Planilha()
    p.carrega_dados(io=param_io, sheet_name=param_sheet_name)

    for i in range(25):
        r = p.le_prox_registro()
        print(r)

    p.recarrega_dados()
    print()
    print('Nova execução pós chamada da função recarrega_dados()')

    print()
    for i in range(25):
        r = p.le_prox_registro()
        print(r)

    print()
    print('Fim teste 02')


## Teste 03 __________________________________________
def teste_03(param_io, param_sheet_name):
    print('Teste 03')
    print('Teste erro - Dicionario sem parametro io.')
    print()

    p = planilha.Planilha()
    p.carrega_dados(io=param_io, sheet_name=param_sheet_name)

    print()
    print('Fim teste 03')


## Teste 04 __________________________________________
def teste_04(param_io, param_sheet_name, param_engine):
    print('Teste 04')
    print('Teste erro - Formato do arquivo de extensão .xlsx e engine diferente de openpyxl.')
    print()

    p = planilha.Planilha()
    p.carrega_dados(io=param_io, sheet_name=param_sheet_name, engine=param_engine)

    print()
    print('Fim teste 04')


## Teste 05 __________________________________________
def teste_05(param_io, param_sheet_name, param_engine):
    print('Teste 05')
    print("Teste erro - Formato do arquivo de extensão .xls e engine diferente de xlrd.")
    print()

    p = planilha.Planilha()
    p.carrega_dados(io=param_io, sheet_name=param_sheet_name, engine=param_engine)

    print()
    print('Fim teste 05')





"""
Gerencia a conexão com um banco de dados.
"""

# TODO: colocar métodos num try. o catch desses try's vai ser feito nas funções que chamarem isso aqui, aqui vai tratar o erro e dar re-raise

import os
import json

from sqlalchemy import create_engine

import blipy.erro as erro


class ConexaoBD():
    """
    Conexão com o banco de dados.

    Qualquer falha na criação dessa conexão dispara uma exceção.
    """

    # tratamento de erro
    # este atributo é um atributo de classe, portanto pode ser alterado pora
    # todas as instâncias dessa classe, se se quiser um tipo de tratamento de
    # erro diferente de imprimir mensagem no console.
    # Para alterá-lo, usar sintaxe "ConexaoBD.e = <novo valor>"
    e = erro.console

    def __init__(self, user, pwd, ip, port, service_name):
        self.__conexao = None

        try:
            cstr = "oracle://" +    user + ":" +                \
                                    pwd + "@" +                 \
                                    ip + ":" +                  \
                                    port + "/?service_name=" +  \
                                    service_name

            engine = create_engine( cstr, 
                                    convert_unicode=True, 
                                    connect_args={"encoding": "UTF-8"})

            self.__conexao = engine.connect()

        except Exception as err:
            self.e._(   "Erro ao conectar com o banco de dados."
                        "\nDetalhes do erro:\n" + str(err))
            raise RuntimeError

    @classmethod
    def from_json(cls):
        """
        Constrói objetos de conexão com o banco a partir de um JSON de
        configuração que esteja no diretório onde o script está sendo
        executado e de uma variável de ambiente que contenha a(s) senha(s)
        do(s) esquema(s).

        O JSON se chamará 'conexoes_bd.json' e terá o seguinte formato:
        {
            "conexoes": [
                {
                    "ordem": <ordem>, 
                    "user": "<usuário>", 
                    "pwd": "<nome da variável de ambiente com a senha>"
                    "ip": "<ip>", 
                    "port": "<número da porta>", 
                    "service_name": "<service_name>" 
                },
                {
                 ....  outra conexão .....
                }
            ]
        }

        O campo ordem do JSON começa com 1.

        A senha de cada conexão será obtida da variável de ambiente cujo nome
        está indicado no parâmetro "pwd".

        Ret:
        :return: retorna uma tupla com um ou mais objetos de conexão com o
        banco de dados, na ordem especificada no campo 'ordem' do JSON.
        """

        param_conexoes = [] 
        try:
            with open("conexoes_bd.json") as f:
                conexoes_json = json.load(f)

                # coloca as conexões lidas do JSON na ordem da tupla de retorno
                indice = 1
                tam = len(conexoes_json["conexoes"])
                terminou = False
                while(True):
                    for conexao in conexoes_json["conexoes"]:
                        if conexao["ordem"] == indice:
                            param_conexoes.append(conexao)
                            indice += 1
                        else:
                            if indice > tam:
                                terminou = True
                            else:
                                continue

                    if terminou:
                        break
        except:
            cls.e._( "Não foi possível ler o arquivo de conexões "
                            "de banco de dados 'conexoes_bd.json'.")
            raise RuntimeError

        conexoes = []
        for conexao in param_conexoes:
            falhou = False
            try:
                senha = os.getenv(conexao["pwd"])
            except KeyError:
                falhou = True

            if (senha is None) or falhou:
                cls.e._( "Não foi possível obter a senha do banco "
                                "de dados de uma variável de ambiente "
                                "para o usuário " + conexao["user"] + ".")
                raise RuntimeError

            try:
                conexoes.append(cls(conexao["user"], 
                                    senha,
                                    conexao["ip"], 
                                    conexao["port"], 
                                    conexao["service_name"]))

            except KeyError as err:
                cls.e._(    "Erro ao conectar com o banco de dados: " 
                            "\nParâmetro de conexao " + str(err) + 
                            " não encontrado.")
                raise RuntimeError

            except Exception as err:
                # tem que ter esse teste aqui pois o construtor já exibirá a
                # mensagem de erro caso haja algum problema com os parâmetros
                if str(err) != "":
                    cls.e._(    "Erro ao conectar com o banco de dados. " 
                                "\nDetalhes do erro:\n" + str(err))
                raise RuntimeError

        return tuple(conexoes)

    def __del__(self):
        if self.__conexao != None:
            try:
                self.__conexao.close()
            except:
                self.e._(   "Erro ao fechar conexão com o banco de dados."
                            "\nDetalhes do erro:\n" + str(err))
                raise RuntimeError



# TODO: precisa desse consulta? não pode ser só o executa? se sim, precisar, então fazê-lo chamar o executa e testar se o sql é mesmo de uma consulta; o try vai ficar no meu executa()
    def consulta(self, sql):
        """
        Executa uma comando de consulta no banco.

        Args:
            sql     : SELECT a ser executado no banco
        Ret:
            Cursor com o resultado do SELECT.
        """

        return self.__conexao.execute(sql)

# TODO: criar métodos insere, atualiza (?)
    def executa(self, sql, commit=False):
        """
        Executa um comando sql no banco.

        Args:
            sql     : sql a ser executado no banco
            commit  : flag indicativa se haverá um commit após a execução do
                      sql ou não 
        Ret:
            Cursor com o resultado do comando sql.
        """

        try:
            if not commit:
                ret = self.__conexao.execute(sql)
            else:
                trans = self.__conexao.begin()
                ret = self.__conexao.execute(sql)
                trans.commit()
        except Exception as err:
            self.e._(   "Erro ao executar o seguinte comando no banco de "
                        "dados:\n" + sql + 
                        "\nDetalhes do erro:\n" + str(err))
            raise RuntimeError

        return ret

    def get_agora(self):
        """
        Retorna o dia e hora atuais do banco.
        """
        return self.consulta("select sysdate from dual").first()[0]

    def tabela_existe(self, tabela):
        """
        Verifica se uma tabela existe no banco.

        Arg:
            tabela  : nome da tabela no banco
        Ret:
            True se tabela existe, False se não existe.
        """
        return self.__conexao.dialect.has_table(self.__conexao, tabela)

    def apaga_registros(self, tabela, condicao=None, commit=False):
        """
        Apaga registros de uma tabela no banco, de acordo com a condição
        informada.

        Args:
            tabela      : nome da tabela no banco
            condicao    : condição WHERE da deleção; se não informado, apaga
                          todas as linhas da tabela
            commit      : flag indicativa se haverá um commit após a execução 
                          da deleção ou não 
        """
        if condicao is not None:
            condicao = " where " + condicao
        else:
            condicao = ""

        sql = "delete from " + tabela + condicao
        try:
            self.executa(sql , commit)
        except Exception as err:
            self.e._(   "Erro ao apagar registros do banco de dados." 
                        "O seguinte comando falhou:\n" + sql + 
                        "\nDetalhes do erro:\n" + str(err))
            raise RuntimeError




"""
Exemplos de uso e casos de teste.
"""

import sys

from blipy.conexao_bd import ConexaoBD
from blipy import tabela_entrada as ti
from blipy import tabela_entrada as ti
from blipy import tabela_saida as to
from blipy import func_transformacao as ft
from blipy import erro as e
from blipy.enum_tipo_col_bd import TpColBD as tp


# TODO: testar algum campo da entrada == null (na leitura da tupla de entrada ele vira None? se sim, meu código continua funcionando?)

def executa_master_job():
    user            = "TESTES_BLIPY"
    pwd             = "4cmSnZDyVhaQx1J"
    ip              = "cluster3-scan"
    port            = "1521"
    service_name    = "DWPROD"

    # conecta com o banco
    try:
        conn = ConexaoBD(user, pwd, ip, port, service_name)
    except:
        sys.exit()

    # cria a tabela de entrada
    abrangencia_in = ti.TabelaEntrada(conn)

    testes = 1023
    if testes & 1:
        # --------------------------------------
        # caso 1: cópia simples de tabelas
        # --------------------------------------

        abrangencia_in.carrega_dados( """
            select 
                ID_ABRANGENCIA, 
                NO_ABRANGENCIA, 
                SN_ATIVO, 
                QT_PONTOS 
            from 
                TESTES_BLIPY.ABRANGENCIA""")

        # cria as colunas da tabela de saída
        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, ft.f_copia), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        # prepara a tabela de saída
        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            # parâmetro de calcula_valor é uma tupla, como só tem um elemento
            # tem que colocar a vírgula no final para construir um objeto do 
            # tipo tupla
            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

            abrangencia_out.grava_registro()

    if testes & 2:
        # --------------------------------------
        # caso 2: preencher campos com NULL
        # há duas formas: passando None como função de transformação ou 
        # passando a tupla (None,) para f_copia
        # --------------------------------------

        abrangencia_in.carrega_dados(
                "TESTES_BLIPY.ABRANGENCIA", 
                ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"])

        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, ft.f_copia), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, None)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (None,) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

            abrangencia_out.grava_registro()

    if testes & 4:
        # --------------------------------------
        # caso 3: transforma entrada num valor fixo
        # --------------------------------------

        abrangencia_in.carrega_dados(
                "ABRANGENCIA", 
                ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"], 
                esquema="TESTES_BLIPY")

        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, ft.ValorFixo("*****")), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.ValorFixo(50))
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor()
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
            abrangencia_out.col["pontos"].calcula_valor()

            abrangencia_out.grava_registro()

    if testes & 8:
        # --------------------------------------
        # caso 4: de/para (inclusive de S/N para 1/0)
        # --------------------------------------

        abrangencia_in.carrega_dados(
            "ABRANGENCIA", 
            ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"], 
            esquema="TESTES_BLIPY")

        # preparar os de/para
        de_para = { "Teste1": "Alterado pelo de/para",
                    "S"     : "1",
                    "N"     : "0"}
        f_de_para = ft.DePara(de_para, copia_se_nao_encontrado=True)

        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, f_de_para), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, f_de_para), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

            abrangencia_out.grava_registro()

    if testes & 16:
        # --------------------------------------
        # caso 5: de/para usando classe específica para S/N
        # --------------------------------------

        abrangencia_in.carrega_dados(
            "ABRANGENCIA", 
            ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"], 
            esquema="TESTES_BLIPY")

        # preparar o de/para
        de_para_nome = {    "Teste1": "Teste1 - Alterado pelo de/para",
                            "Teste2": "Teste2 - Alterado pelo de/para"}
        f_de_para_nome = ft.DePara(de_para_nome, copia_se_nao_encontrado=True)

        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, f_de_para_nome), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, 
                                                ft.DeParaSN(val_int=False)), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

            abrangencia_out.grava_registro()

    if testes & 32:
        # --------------------------------------
        # caso 6: somatório e média
        # --------------------------------------

        abrangencia_in.carrega_dados(
            "ABRANGENCIA", 
            ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"], 
            esquema="TESTES_BLIPY")
        
        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, ft.f_copia), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_somatorio)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        # calcula o somatário
        valor_anterior = 0
        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )

            # cada registro é o somatório dele com a soma dos anteriores
            # abrangencia_out.col["pontos"].calcula_valor( (valor_anterior, registro[3]) )
            # valor_anterior = abrangencia_out.col["pontos"].valor

            # cada registro é o somatário dele com a anterior
            abrangencia_out.col["pontos"].calcula_valor( (valor_anterior, registro[3] ))
            valor_anterior = registro[3]

            abrangencia_out.grava_registro()


        # ------------------
        # calcula a média

        abrangencia_in.recarrega_dados()

        # altera a função de transformação para "média"
        abrangencia_out.col["pontos"] = to.Coluna(  "QT_PONTOS", 
                                                    tp.NUMBER, 
                                                    ft.f_media)

        # calcula a média
        valor_anterior = 0
        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )

            # cada registro é a média dele com a anterior
            abrangencia_out.col["pontos"].calcula_valor( (valor_anterior, registro[3] ) )
            valor_anterior = registro[3]

            abrangencia_out.grava_registro()

    if testes & 64:
        # --------------------------------------
        # caso 7: data (e hora) atuais
        # --------------------------------------

        abrangencia_in.carrega_dados(
                "ABRANGENCIA", 
                ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"], 
                esquema="TESTES_BLIPY")
        
        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, ft.f_copia), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        nr_registro = 0
        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            if nr_registro == 0:
                abrangencia_out.col["data_atualizacao"] = \
                    to.Coluna("DT_ATUALIZACAO", tp.DATE, 
                            ft.Agora(conn))
            elif nr_registro == 1:
                abrangencia_out.col["data_atualizacao"] = \
                    to.Coluna("DT_ATUALIZACAO", tp.DATE, 
                            ft.Agora(conn, so_data=True))
            elif nr_registro == 2:
                abrangencia_out.col["data_atualizacao"] = \
                    to.Coluna("DT_ATUALIZACAO", tp.DATE, 
                            ft.Agora())
            elif nr_registro == 3:
                abrangencia_out.col["data_atualizacao"] = \
                    to.Coluna("DT_ATUALIZACAO", tp.DATE, 
                            ft.Agora(so_data=True))

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )
            abrangencia_out.col["data_atualizacao"].calcula_valor()

            abrangencia_out.grava_registro()

            nr_registro += 1

    if testes & 128:
        # --------------------------------------
        # caso 8: força erro para emitir mensagem formatada
        # --------------------------------------

        #+++++++++++++
        # erro 1
        #+++++++++++++

        abrangencia_in.carrega_dados(
                "ABRANGENCIA", 
                ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"], 
                esquema="TESTES_BLIPY")
 #
        de_para = { "Teste1": "Alterado pelo de/para",
                    "S"     : "1",
                    "N"     : "0"}
        f_de_para = ft.DePara(de_para, copia_se_nao_encontrado=True)

        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, f_de_para), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, f_de_para), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            try:
                abrangencia_out.col["id"].calcula_valor(     (registro[0],) )

                # código forçando um erro (não poderia haver mais de uma entrada)
                abrangencia_out.col["nome"].calcula_valor(   (registro[1], registro[2]) )

                abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
                abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

                abrangencia_out.grava_registro()

            except:
                e.console._("Erro no master job. Abortando.")
                break


        #+++++++++++++
        # erro 2
        #+++++++++++++

        abrangencia_in.recarrega_dados()

        de_para = { "Teste1": "Alterado pelo de/para",
                    "S"     : "1",
                    "N"     : "0"}
        f_de_para = ft.DePara(de_para, copia_se_nao_encontrado=True)

        col = {
            # código forçando um erro (ainda não há tratamento para tipo BOOL)
            "id"        : to.Coluna("ID_ABRANG", tp.BOOL, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, f_de_para), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, f_de_para), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            try:
                abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
                abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
                abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
                abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

                abrangencia_out.grava_registro()

            except:
                e.console._("Erro no master job. Abortando.")
                break

    if testes & 256:
        # --------------------------------------
        # caso 9: profiling de performance
        # --------------------------------------

        abrangencia_in.carrega_dados(
                "TESTES_BLIPY.ABRANGENCIA", 
                ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"])

        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : to.Coluna("NO_ABRANG", tp.STRING, ft.f_copia), 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, None)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        # habilita o profiling de performance
        abrangencia_out.habilita_profiling_performance()

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )
            abrangencia_out.col["ativo"].calcula_valor(  (None,) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

            abrangencia_out.grava_registro()

    if testes & 512:
        # --------------------------------------
        # caso 10: trasnformação custom
        # --------------------------------------

        abrangencia_in.carrega_dados(
                "TESTES_BLIPY.ABRANGENCIA", 
                ["ID_ABRANGENCIA", "NO_ABRANGENCIA", "SN_ATIVO", "QT_PONTOS"])

        nome_invertido = to.Coluna("NO_ABRANG", tp.STRING, InverteString())
        nome_copiado   = to.Coluna("NO_ABRANG", tp.STRING, ft.f_copia) 
        col = {
            "id"        : to.Coluna("ID_ABRANG", tp.NUMBER, ft.f_copia), 
            "nome"      : nome_invertido, 
            "ativo"     : to.Coluna("SN_ATIVO",  tp.STRING, ft.f_copia), 
            "pontos"    : to.Coluna("QT_PONTOS", tp.NUMBER, ft.f_copia)
            }

        abrangencia_out = to.TabelaSaida("ZZ_TESTE", col, conn)

        while True:
            registro = abrangencia_in.le_prox_registro()
            if registro is None:
                break

            abrangencia_out.col["id"].calcula_valor(     (registro[0],) )

            if registro[1] == "Teste4":
                col["nome" ] = nome_copiado
            abrangencia_out.col["nome"].calcula_valor(   (registro[1],) )

            abrangencia_out.col["ativo"].calcula_valor(  (registro[2],) )
            abrangencia_out.col["pontos"].calcula_valor( (registro[3],) )

            abrangencia_out.grava_registro()

            if registro[1] == "Teste4":
                col["nome" ] = nome_invertido

    if testes & 1024:
        # --------------------------------------
        # caso 11: trasnformação DeParaChar
        # --------------------------------------

        # cria a tabela de entrada
        tab_entrada = ti.TabelaEntrada(conn)
        tab_entrada.carrega_dados("""
                SELECT 
                        ID_REGISTRO,
                        DS_TEXTO

                FROM    TAB_IN_TESTE_DEPARACHAR
            """)

        # cria as colunas da tabela de saída
        col = {
            "ID_REGISTRO": to.Coluna("ID_REGISTRO", tp.NUMBER, ft.f_copia),
            "DS_TEXTO": to.Coluna("DS_TEXTO", tp.STRING, ft.DeParaChar({"'": "''",'"': "*",}))
        }

        tab_saida = to.TabelaSaida("TESTES_BLIPY.TAB_OUT_TESTE_DEPARACHAR", col, conn)

        while True:
            registro = tab_entrada.le_prox_registro()

            if registro is None:
                break

            tab_saida.col["ID_REGISTRO"].calcula_valor((registro[0],))
            tab_saida.col["DS_TEXTO"].calcula_valor((registro[1],))

            tab_saida.grava_registro()

class InverteString():
    """
    Inverte uma string de entrada.
    """

    def transforma(self, entradas):
        """
        Inverte a string de entrada.

        Args:
            entradas   : tupla com o valor a string a ser invertida; 
                         só pode haver um elemento na tupla
        Ret:
            string de entrada invertida
        """

        if len(entradas) != 1:
            raise RuntimeError(
                "Não pode haver mais de um dado de entrada.")

        # não funciona pra utf-8!
        return entradas[0][::-1]




"""
Funções para facilitar a implementação de um ETL, subindo o nível de abstração
para um job de carga de dados.

Se parâmetro '-v' (de verbose) for passado na linha de comando ao executar o
script de carga, a quantidade de registros lidos e gravados será impressa no
console.
"""

import sys
from datetime import datetime

import blipy.tabela_entrada as te
import blipy.tabela_saida as ts
import blipy.func_transformacao as ft

from enum import Enum, auto


# Tipos de estratégias de gravação no banco
class TpEstrategia(Enum):
    DELETE  = auto()
    INSERT  = auto()


class Job():
    def __init__(self, nome_job):
        self.__verbose = False
        if len(sys.argv) == 2:
            if sys.argv[1] == "-v":
                self.__verbose = True
 
        self.__nome = nome_job
 
        print()
        print("====== Job " + self.__nome + " iniciado " + "=======")
        print("-----> Horário de início:  " +  \
                str(datetime.now().replace(microsecond=0)))

    def __del__(self):
        print("====== Job " + self.__nome + " finalizado " + "=====")
        print("-----> Horário de término: " +  \
                str(datetime.now().replace(microsecond=0)))

    def importa_tabela_por_sql(self, conn_entrada, conn_saida, sql_entrada,
            nome_tabela_saida, cols_saida, estrategia=TpEstrategia.DELETE):
        """
        Importação do resultado de uma consulta sql para o banco de dados. A
        tabela de destino é limpa e carregada de novo (considera-se que são poucas
        linhas). Qualquer erro dispara uma exceção.

        Args:
        :param conn_entrada: conexão com o esquema de entrada de dados (geralmente
        o staging)
        :param conn_saida: conexão com o esqumea de saída (geralmente a produção)
        :param sql_entrada: consulta sql que irá gerar os dados de entrada
        :param nome_tabela_saida: nome da tabela de saida
        :param cols_saida: lista das colunas que serão gravadas na tabela 
        de saida, com o nome da coluna, seu tipo e a função de transformanção
        a ser aplicada (opcional; se não informado, faz só uma cópia do dado)
        :param estrategia: estratégia de gravação que será utilizada (enum
        TpEstrategia)
        """

        tab_entrada = te.TabelaEntrada(conn_entrada)
        tab_entrada.carrega_dados(sql_entrada)

        self.__grava_tabela(    tab_entrada, 
                                conn_saida, 
                                nome_tabela_saida, 
                                cols_saida, 
                                estrategia)

    def importa_tabela_por_nome(self, conn_entrada, conn_saida, 
            nome_tabela_entrada, nome_tabela_saida, cols_entrada, cols_saida, 
            estrategia=TpEstrategia.DELETE):
        """
        Importação de uma tabela para o banco de dados. A tabela de destino é limpa
        e carregada de novo (considera-se que são poucas linhas). Qualquer erro
        dispara uma exceção.

        Args:
        :param conn_entrada: conexão com o esquema de entrada de dados (geralmente
        o staging)
        :param conn_saida: conexão com o esqumea de saída (geralmente a produção)
        :param nome_tabela_entrada: nome da tabela de entrada
        :param nome_tabela_saida: nome da tabela de saida
        :param cols_entrada: lista dos nomes das colunas que serão buscadas na 
        tabela de entrada
        :param cols_saida: lista das colunas que serão gravadas na tabela 
        de saida, com o nome da coluna, seu tipo e a função de transformanção
        a ser aplicada (opcional; se não informado, faz só uma cópia do dado)
        :param estrategia: estratégia de gravação que será utilizada (enum
        TpEstrategia)
        """

        tab_entrada = te.TabelaEntrada(conn_entrada)
        tab_entrada.carrega_dados(nome_tabela_entrada, cols_entrada)

        self.__grava_tabela(    tab_entrada, 
                                conn_saida, 
                                nome_tabela_saida, 
                                cols_saida, 
                                estrategia)

    def __grava_tabela(self, 
            tab_entrada, 
            conn_saida, 
            nome_tabela_saida, 
            cols_saida, 
            estrategia):
        """
        Grava uma tabela no banco de dados.

        Args:
        :param tab_entrada: a tabela de entrada dos dados
        :param conn_saida: conexão com o esqumea de saída (geralmente a produção)
        :param nome_tabela_saida: nome da tabela de saida
        :param cols_saida: lista das colunas que serão gravadas na tabela 
        de saida, com o nome da coluna, seu tipo e a função de transformanção
        a ser aplicada (opcional; se não informado, faz só uma cópia do dado)
        :param estrategia: estratégia de gravação que será utilizada (enum
        TpEstrategia)
        """

        cols = {}
        for item in cols_saida:
            if len(item) == 2:
                # função de transformanção não informada, faz só uma cópia do dado
                cols[item[0]] = ts.Coluna(item[0], item[1], ft.f_copia)
            else:
                # usa a função de transformanção informada
                cols[item[0]] = ts.Coluna(item[0], item[1], item[2])

        tab_saida = ts.TabelaSaida(nome_tabela_saida, cols, conn_saida)

        if estrategia == TpEstrategia.DELETE:
            conn_saida.apaga_registros(nome_tabela_saida)
        elif estrategia == TpEstrategia.INSERT:
            pass
        else:
            raise NotImplementedError

        qtd_registros = 0
        while True:
            registro = tab_entrada.le_prox_registro()
            if registro is None:
                break

            i = 0
            for k in cols.keys():
                tab_saida.col[k].calcula_valor( (registro[i],) )
                i += 1

            tab_saida.grava_registro()
            qtd_registros += 1
 
        if self.__verbose:
            if tab_entrada.nome == "":
                # a consulta de entrada não leu uma só tabela, mas um
                # select provavelmente com joins de tabelas
                print(  str(qtd_registros) +  \
                        " \t\tregistros lidos dos dados de entrada e" 
                        " salvos na tabela " +  \
                        nome_tabela_saida)
            else:
                print(  str(qtd_registros) +  \
                        " \t\tregistros lidos da tabela " + tab_entrada.nome + \
                        " e salvos na tabela " +  \
                        nome_tabela_saida)





"""
Gerencia a carga dos dados de uma tabela do banco.
"""

# módulo usado para  implementar method overloading
# ATENÇÃO: este módulo *não é* thread safe
# (https://stackoverflow.com/questions/6434482/python-function-overloading)
# ATENÇÃO: métodos que usam o decorator dispatch e que contenham argumentos
# nomeados com valores default: se o valor default for alterado na chamada do
# método, precisa ser por meio de argumento nomeado, não pode se basear na
# posição do argumento
# (https://stackoverflow.com/questions/54132640/how-to-use-default-parameter-with-multipledispatch)
from multipledispatch import dispatch

import blipy.erro as erro


class TabelaEntrada():
    """
    Tabela a ser lida do banco de dados.
    """

    # tratamento de erro
    # este atributo é um atributo de classe, portanto pode ser alterado pora
    # todas as instâncias dessa classe, se se quiser um tipo de tratamento de
    # erro diferente de imprimir mensagem no console.
    # Para alterá-lo, usar sintaxe "TabelaEntrada.e = <novo valor>"
    e = erro.console

    def __init__(self, conexao_bd):
        """
        Args:
            conexao_bd  : conexão com o banco de dados
        """

        self.__sql          = ""
        self.__conexao_bd   = conexao_bd
        self.__cursor       = None
        self.__nome_tabela  = ""

    # getter; instancia.nome vai retornar __nome_tabela
    @property
    def nome(self):
        return self.__nome_tabela
    @nome.setter
    def nome(self, nome):
        raise NotImplementedError

    @dispatch(str)
    def carrega_dados(self, sql):
        """
        Carrega a tabela com os dados do banco.

        Args:
            sql : sql no banco cujo resultado será a tabela de entrada. Pode
                  ser apenas um filtro numa tabela ou mesmo joins entre tabelas
        Ret:
            um iterável com os registros da tabela do banco.
        """

        self.__sql = sql

        self.__cursor = None
        return self.__carrega_tabela()

    # como usa o decorator dispatch, o código que chama esse método tem que 
    # nomear o arqgumento 'esquema' se quiser alterá-lo
    # (https://stackoverflow.com/questions/54132640/how-to-use-default-parameter-with-multipledispatch)
    @dispatch(str, list, esquema=str)
    def carrega_dados(self, tabela, colunas, esquema=""):
        """
        Carrega a tabela com os dados do banco.

        Args:
            tabela  : nome da tabela no banco
            colunas : lista com os nomes das colunas da tabela que ser quer
                      carregar
            esquema : nome do esquema
        Ret:
            um iterável com os registros da tabela do banco.
        """

        self.__nome_tabela = tabela

        cols = ""
        for col in colunas:
            cols += col + ", "
        cols = cols[:len(cols)-2]

        if esquema != "":
            esquema += "."
        self.__sql = "select " + cols + " from " + esquema + tabela

        self.__cursor = None
        return self.__carrega_tabela()

    def recarrega_dados(self):
        """
        Rearrega a tabela com os dados do banco.

        Ret:
            um iterável com os registros da tabela do banco ou uma exceção
            se a tabela ainda não foi carregada.
        """

        if self.__sql == "":
            self.e._("A tabela não pode ser recarregada pois ainda não "
                     "foi carregada pela primeira vez.")
            raise RuntimeError

        self.__cursor = None
        return self.__carrega_tabela()

    def __carrega_tabela(self):
        """
        Carrega a tabela com os dados do banco.

        Ret:
            um iterável com os registros da tabela do banco.
        """

        if self.__cursor is None:
            self.__cursor = self.__conexao_bd.executa(self.__sql)

        for row in self.__cursor:
            # é necessário usar yield pois o cursor do sql não é exposto para
            # fora dessa classe, e le_prox_registro (que chama este método
            # aqui) deve ser chamado num loop no job de carga
            yield row

    def le_prox_registro(self):
        """
        Lê a próximo registro da tabela.

        Ret:
            tupla com o registro lido ou None se não houver mais registros a
            serem lidos.
        """

        registro = self.__carrega_tabela()
        try:
            ret = next(registro)
        except StopIteration:
            ret = None

        return ret





"""
Módulo para tratamento de erros.
"""

import inspect

class TpSaida():
    CONSOLE     = 1
    # valores dos próximos têm que ser o valor anterior * 2, Por exemplo:
    # (isso vai permitir mais de uma saída ao mesmo tempo)
    # EMAIL     = 2
    # TEAMS     = 4
    # BD        = 8

class TratamentoErro():
    """
    Classe para emissão de mensagens de erro formatadas. Para ser usado dentro
    de um bloco try..except, por exemplo.
    """

    def __init__(self, saida=TpSaida.CONSOLE):
        """
        Args:
            saida:  : indica onde a mensagem de erro será exibida.
                       Atualmente, única opção é exibir o erro no console.
        """

        self.__saida = saida

    def _(self, msg):
        """
        Emite a mensagem de erro formatada na saída definida para esta
        instância.

        Args:
            msg:  a mensagem de erro
        """

        #  pega objeto que chamou essse método aqui
        chamador = inspect.stack()[1]

        if "self" in chamador.frame.f_locals:
            obj = " de " + str(type(chamador.frame.f_locals["self"])) + ":"

            # pega o nome do método que  chamou o tratamento do erro
            # https://stackoverflow.com/questions/5067604/determine-function-name-from-within-that-function-without-using-traceback
            func = chamador[3]
        elif "__name__" in chamador.frame.f_locals:
            obj = ":"
            func = chamador.frame.f_locals["__name__"]
        else:
            obj = ":"
            func = "não foi possível identificar a função"

        if self.__saida & TpSaida.CONSOLE:
            print()
            print("============= ERRO ================")
            print(  "Ocorreu um erro " 
                    f"na função <{func}>{obj}" )
            print(msg)
            print("===================================")
            print()
        else:
            raise NotImplementedError

# Instância criada por conveniência para uso mais comum desta classe
console = TratamentoErro(TpSaida.CONSOLE)





import sys
sys.path.append('..')

from blipy.conexao_bd import ConexaoBD

from blipy.job import Job
from blipy.enum_tipo_col_bd import TpColBD as tp


if __name__ == "__main__":
    job = Job("Teste Carga")

    try:
        conn_stg, conn_prd, conn_corp = ConexaoBD.from_json()
    except:
        sys.exit()

    print(conn_stg)
    print(conn_prd)
    print(conn_corp)

    # dimensão LOCAL_HOSPEDAGEM
    cols_entrada = ["ID_LOCAL_HOSPEDAGEM",
                    "NO_LOCAL_HOSPEDAGEM"]
    cols_saida = [  ["ID_LOCAL_HOSPEDAGEM", tp.NUMBER],
                    ["NO_LOCAL_HOSPEDAGEM", tp.STRING]]
    job.importa_tabela_por_nome(   
            conn_stg, 
            conn_prd, 
            "MVW_LOCAL_HOSPEDAGEM", 
            "LOCAL_HOSPEDAGEM",
            cols_entrada, 
            cols_saida)





"""
Gerencia a gravação de dados numa tabela no banco.
"""

# TODO: ajustar tratamento de erro para tratamento próprio

import blipy.erro as erro
from blipy.profiling import ProfilingPerformance
from blipy.enum_tipo_col_bd import TpColBD as tp


class TabelaSaida(ProfilingPerformance):
    """
    Tabela a ser gravada no banco de dados.
    """

    # tratamento de erro
    # este atributo é um atributo de classe, portanto pode ser alterado pora
    # todas as instâncias dessa classe, se se quiser um tipo de tratamento de
    # erro diferente de imprimir mensagem no console.
    # Para alterá-lo, usar sintaxe "TabelaSaida.e = <novo valor>"
    e = erro.console

    def __init__(self, nome, colunas, conexao_bd):
        """
        Args:
            nome        : nome da tabela no banco
            colunas     : dict com objetos do classe Coluna representando as
                          colunas do banco
            conexao_bd  : conexão com o banco de dados
        """

        self.__nome         = nome
        self.__conexao_bd   = conexao_bd
        self.col            = colunas

        self._profiling_on = False


    def habilita_profiling_performance(self, path_csv=""):
        """
        Habilita o registro de performance para essa classe em si e para todas
        as colunas que compõem a tabela.
        """
        
        # habilita profiling pra todas as colunas que compõem a tabela
        for key in list(self.col):
            self.col[key].habilita_profiling_performance(path_csv)

        self._profiling_on = True

# TODO: criar parâmetro para estratégia de gravacao (limpa e grava tudo, insert, update etc.)
    def grava_registro(self):
        """
        Grava os dados da tabela no banco.

        Um erro de banco disparará uma exceção.
        """

        if self._profiling_on:
            self._inicia_timer()

        cols = ""
        for coluna in self.col:
            cols += self.col[coluna].nome + ", "
        cols = cols[:len(cols)-2]

        # TODO: tratar demais tipos possíveis do banco (como bool)
        values = ""
        for coluna in self.col:
            if self.col[coluna].valor is None:
                values += "NULL"
            else:
                if self.col[coluna].tipo == tp.STRING:
                    values += "'" + self.col[coluna].valor + "'"
                elif self.col[coluna].tipo == tp.INT       or \
                     self.col[coluna].tipo == tp.NUMBER:
                    values += str(self.col[coluna].valor)

                elif self.col[coluna].tipo == tp.DATE:
                    values += "TO_DATE('" + str(self.col[coluna].valor) + \
                        "', 'yyyy-mm-dd hh24:mi:ss')"
                else:
                    self.e._(   "Tipo de dado inválido para a coluna " \
                                + self.col[coluna].nome + " da tabela " + \
                                self.__nome + ".")
                    if self._profiling_on:
                        self._finaliza_timer()
# TODO: disparar uma exceção específica pro master job saber se continua ou não (?)
                    raise RuntimeError

            values += ", "
        values = values[:len(values)-2]

        # TODO: implementar os outros estratégias além do insert: insert_update, update_insert, update
        self.__conexao_bd.executa(
                "insert into " 
                + self.__nome + 
                " (" + cols + ") " + 
                " values ( " + values +
                ") "
            )

        if self._profiling_on:
            self._finaliza_timer()

class Coluna(ProfilingPerformance):
    """
    Uma coluna de uma tabela do banco de dados.
    """

# TODO: parâmetro 'tipo' seria interessante se fosse pra fazer um cast automático do resultado da ft para o valor da coluna
    def __init__(self, nome, tipo, func_transf):
        """
        Args:
            nome        : nome da coluna no banco de dados
            tipo        : tipo da coluna no banco de dados. É um elemento do 
                          enum tipo_col_bd
            func_transf : função de transformação que será aplicada para obter
                          o valor da coluna. Se for None, valor da coluna no 
                          banco de dados será NULL
        """
        self.nome           = nome
        self.tipo           = tipo
        self.__func_transf  = func_transf
        self.valor          = None

        # tratamento de erro
        # este atributo é público, portanto pode ser alterado por quem usa 
        # esta classe se quiser um tipo de tratamento de erro diferente de 
        # imprimir mensagem no console
        self.e = erro.console

        self._profiling_on = False

    def calcula_valor(self, entradas=None):
        """
        Calcula o valor da coluna de acordo com sua função de transformação.
        Se a função de transformação for None, valor será NULL (None).

        Arg:
            entradas    : tupla de valores a ser passado para a função de
                          transformação
        """

        if self._profiling_on:
            self._inicia_timer()

        if self.__func_transf is None:
            self.valor = None
        else:
            try:
                self.valor = self.__func_transf.transforma(entradas)
            except Exception as err:
                self.e._(
                    f"Erro ao calcular o valor da coluna {self.nome}." + 
                    "\n" + str(err))
                if self._profiling_on:
                    self._finaliza_timer()
                raise RuntimeError

        if self._profiling_on:
            self._finaliza_timer()





from setuptools import setup, find_packages

VERSION = '2.1.2' 
DESCRIPTION = 'Framework para rotinas de ETL'
LONG_DESCRIPTION = 'Framework para simplificar o desenvolvimento de rotinas python, especialmente rotinas de ETL'

# Setting up
setup(
        name="stnblipy", 
        version=VERSION,
        author="Secretaria do Tesouro Nacional",
        # author_email="",
        license="GPL v2",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        packages=find_packages(),
        install_requires=[
            "sqlalchemy==1.4.15", 
            "cx-Oracle==7.1.3", 
            "multipledispatch==0.6.0 ", 
            "pandas==1.1.5", 
            "openpyxl==3.0.10" ], 
        keywords=['python', 'ETL', 'framework'],
        classifiers= [
            "Programming Language :: Python :: 3.7",
        ]
)





"""
Enumeradores para tipos de colunas no banco de dados. 
"""

from enum import Enum, auto

# TODO: criar demais tipos possíveis do banco (como bool)

# Tipos de colunas no banco de dados
class TpColBD(Enum):
    STRING  = auto()
    INT     = auto()
    NUMBER  = auto()
    DATE    = auto()
    BOOL    = auto()




"""
Gerencia a carga dos dados de uma planilha do Excel.
"""

# módulo usado para  implementar method overloading
# ATENÇÃO: este módulo *não é* thread safe
# (https://stackoverflow.com/questions/6434482/python-function-overloading)
# ATENÇÃO: métodos que usam o decorator dispatch e que contenham argumentos
# nomeados com valores default: se o valor default for alterado na chamada do
# método, precisa ser por meio de argumento nomeado, não pode se basear na
# posição do argumento
# (https://stackoverflow.com/questions/54132640/how-to-use-default-parameter-with-multipledispatch)

import pandas as pd

import blipy.erro as erro


class Planilha():
    """
    Planilha a ser carregada no banco de dados.
    """

    # tratamento de erro
    # este atributo é um atributo de classe, portanto pode ser alterado pora
    # todas as instâncias dessa classe, se se quiser um tipo de tratamento de
    # erro diferente de imprimir mensagem no console.
    # Para alterá-lo, usar sintaxe "Planilha.e = <novo valor>"
    e = erro.console

    def carrega_dados(self, 
            io=None, sheet_name=0, header=0, skipfooter=0, 
            names=None, index_col=None, usecols=None,
            skiprows=None, nrows=None, na_values=None, engine='openpyxl'):
        """
        Lê uma aba de uma planilha de Excel e carrega em um Data Frame do
        Pandas.

        Referência: https://pandas.pydata.org/docs/reference/api/pandas.read_excel.html

        Args:
        :param io:         Caminho (path) do arquivo, com nome do arquivo incluso ou uma URL
        :param sheet_name: Aba ou guia do Excel, pode ser o nome ou número 0-indexed
        :param header:     Linhas de cabeçalho a serem ignoradas, 0-indexed
        :param skipfooter: Linhas de rodapé (final arquivo) a serem ignoradas, 0-indexed.
        :param names:      Lista de nomes de colunas. Se arquivo não contém header, deve-se obrigatoriamente setar header=None
        :param index_col:  int, list of int, default None - Coluna com o índice do
        :param usecols:    Seleção de colunas. Tipos de seleção: 'A,B,C', [0,1,2], ['ID','Coluna_1','Coluna_4']
        :param skiprows:   Linhas a serem ignoradas no início do arquivo, 0-indexed.
        :param nrows:      Número de linhas carregadas.
        :param na_values:  Lista de strings a serem consideradas como dados não disponívels NaN
        :param engine:     O padrão é openpyxl. Existem outras para formatos antigos ex .odf

        Ret:
        :return: self.__tabela_data_frame
        """

        self.__tabela_data_frame = None  # Data Frame da planilha carregada
        self.__cursor            = None  # Iterável para função le_prox_registro()

        # Verificações de possíveis erros
        if io is None:
            self.e._("O argumento 'io' é o mínimo para funcionamento da função. \n"
                     "Este é o caminho (path) para a planilha com o nome do arquivo incluso ao final e com a extensão .xlsx ou .xls. \n"
                     "Formato esperado: C:/.../planilha.xlsx")
            raise RuntimeError

        if io.endswith('.xlsx') and engine != 'openpyxl':
            self.e._("Parâmetro 'engine' não adequado para o tipo de arquivo.\n"
                     "O argumento 'io' caminho (path) para a planilha deve terminar com: .xlsx ou .xls.\n"
                     "Para cada formato de arquivo existe uma engine adequada.\n"
                     "Engine 'openpyxl' para .xlsx e 'xlrd' para .xls.\n"
                     "Exemplos de parâmetro 'io' esperados:\n"
                     "\tC:/.../planilha.xlsx\n"
                     "\tC:/.../planilha.xls\n")
            raise RuntimeError

        if io.endswith('.xls') and engine != 'xlrd':
            self.e._("Parâmetro 'engine' não adequado para o tipo de arquivo.\n"
                     "O argumento 'io' caminho (path) para a planilha deve terminar com: .xlsx ou .xls.\n"
                     "Para cada formato de arquivo existe uma engine adequada.\n"
                     "Engine 'openpyxl' para .xlsx e 'xlrd' para .xls.\n"
                     "Exemplos de parâmetro 'io' esperados:\n"
                     "\tC:/.../planilha.xlsx\n"
                     "\tC:/.../planilha.xls\n")
            raise RuntimeError

        if not io.endswith(('.xlsx','.xls')):
            self.e._("O argumento 'io' caminho (path) para a planilha deve terminar com: .xlsx ou .xls.\n"
                     "Para cada formato de arquivo existe uma engine adequada.\n"
                     "Engine 'openpyxl' para .xlsx e 'xlrd' para .xls.\n"
                     "Exemplos de parâmetro 'io' esperados:\n"
                     "\tC:/.../planilha.xlsx\n"
                     "\tC:/.../planilha.xls\n")
            raise RuntimeError

        # Carregamento da tabela para a memória
        self.__tabela_data_frame = pd.read_excel(io=io,sheet_name=sheet_name, 
                header=header, skipfooter=skipfooter,
                names=names, index_col=index_col, usecols=usecols, 
                skiprows=skiprows, nrows=nrows, na_values=na_values, 
                engine=engine)


    def recarrega_dados(self):
        """
        Faz o reset do cursor para o ponto inicial a partir do data frame
        carregado.

        Ret:
        Iterável com os registros de una tabela do Excel
        se a tabela ainda não foi carregada em um data frame retorna erro.
        """
        if self.__tabela_data_frame is None:
            self.e._("A planilha não pode ser recarregada pois ainda não "
                     "foi carregada pela primeira vez.\n"
                     "Necessário executar a função carrega_dados() primeiro.")
            raise RuntimeError

        # Reset do cursor para o index 0.
        self.__cursor = None
        self.__cursor = self.__tabela_data_frame.itertuples()


    def le_prox_registro(self):
        """
        Lê a próximo registro da tabela.

        Ret:
            tupla com o registro lido ou None se não houver mais registros a
            serem lidos.
        """

        if self.__cursor is None:
            self.__cursor = self.__tabela_data_frame.itertuples()

        try:
            registro = next(self.__cursor)
            registro = tuple(registro)
        except StopIteration:
            registro = None

        return registro



"""
Funções de transformação a serem usadas nas cargas ("T" do ETL).
"""

import html
import re
from datetime import date, datetime


"""
Módulo com as classes de transformação de dados ("T" do ETL)
"""

class Copia():
    """
    Cópia simples de dados, sem transformações ou nenhum tipo de alteração.
    """

    def transforma(self, entradas):
        """
        Copia a entrada, sem nenhuma transformação.

        Args:
            entradas   : tupla com o valor a ser copiado; como não faz
                         transformação alguma, só pode haver um elemento na
                         tupla
        Ret:
            cópia do valor de entrada
        """

        if len(entradas) != 1:
            raise RuntimeError(
                "Não pode haver mais de um dado de entrada para cópia.")

        return entradas[0]

class ValorFixo():
    """
    Transforma a entrada num valor fixo.
    """

    def __init__(self, valor):
        self.__valor = valor

    def transforma(self, entradas):
        """
        Retorna um valor fixo, definido na instanciação do objeto. Para valores
        NULL em banco de dados, informar tipo de dados None do Python.

        Args:
            entradas   : não é necessário, mantido apenas para consistência da 
                         interface
        Ret:
            Valor definido na instanciação da classe
        """

        return self.__valor

class DePara():
    """
    Faz um de/para dos valores de entrada.
    """

    def __init__(self, de_para, copia_se_nao_encontrado=True):
        """
        Args:
            de_para                 : dict com o de/para desejado
            copia_se_nao_encontrado : se True, copia valor da entrada para a
                                      saída caso valor de entrada não seja
                                      encontrado no de/para; se False, dispara
                                      exceção
        """
        self.__de_para = de_para
        self.__copia_se_nao_encontrado = copia_se_nao_encontrado

# TODO: criar outras condições se não achar o valor no de/para: preenche null ou um valor específico; tem que mudar o método transforma() abaixo e ver quais impactos na classe filha deparasn

    def transforma(self, entradas):
        """
        Retorna um de/para dos valores de entrada.

        Args:
            entradas   : tupla com o valor a ser transformado; como só
                         transforma um valor em outro, só pode haver um
                         elemento na tupla
        Ret:
            De/para da entrada.
        """

        if len(entradas) != 1:
            raise RuntimeError(
                "Não pode haver mais de um dado de entrada para um de/para.")

        for item in entradas:
            if item in list(self.__de_para):
                return self.__de_para[item]

        if self.__copia_se_nao_encontrado:
            return entradas[0]
        else:
# TODO: disparar uma exceção específica pra master job decidir se quer continuar ou não se não achar de/para
            #  não achou a key no de_para informado no construtor
            raise RuntimeError(
                "Impossível fazer a transformação, valor de de/para não encontrado.")

class DeParaSN(DePara):
    """
    Faz um de/para de campos S/N para 1/0 e vice-versa. Considera também
    a língua (por exemplo, Y/N ao invés de S/N).
    """

    def __init__(self, 
            copia_se_nao_encontrado=True, 
            inverte=False, 
            val_int=True, 
            lingua="pt"):
        """
        Args:
            copia_se_nao_encontrado  : dispara execeção de erro caso valor não
                                      seja encontrado no de/para; se False,
                                      copia valor da entrada para a saída
            inverte                 : se de/para é de S/N para 1/0 ou de 1/0 
                                      para S/N
            val_int                 : se valor 1/0 deve ser um inteiro ou
                                      string
            lingua                  : "pt" ou "en"
        """

        if lingua == "pt" and val_int and not inverte:
            de_para = {"S": 1, "N": 0}
        elif lingua == "pt" and not val_int and not inverte: 
            de_para = {"S": "1", "N": "0"}

        elif lingua == "pt" and val_int and inverte: 
            de_para = {1: "S", 0: "N"}
        elif lingua == "pt" and not val_int and inverte: 
            de_para = {"1": "S", "0": "N"}

        elif lingua == "en" and val_int and not inverte: 
            de_para = {"Y": 1, "N": 0}
        elif lingua == "en" and not val_int and not inverte: 
            de_para = {"Y": "1", "N": "0"}

        elif lingua == "en" and val_int and inverte: 
            de_para = {1: "Y", 0: "N"}
        elif lingua == "en" and not val_int and inverte: 
            de_para = {"1": "Y", "0": "N"}

        super().__init__(de_para, copia_se_nao_encontrado)

class DeParaChar():
    """
    Faz um de/para de um ou mais caracteres em um texto.
    """

    def __init__(self, de_para = None):
        """
        Args:
             de_para  : dict com o(s) de/para de caracteres desejados
        """
        self.__de_para = de_para

    def transforma(self, entradas):
        """
        Retorna um texto com um conjunto de caracteres transformados a partir de
        um dict de/para.

        Args:
            entradas : tupla contendo o texto de entrada a ser transformado
        """

        if self.__de_para is None:
            raise RuntimeError(
                "Impossível fazer a transformação, dicionário de/para não encontrado.")

        texto = entradas[0]
        for chave in self.__de_para.keys():
            texto = texto.replace(chave, self.__de_para[chave])

        return texto

class Somatorio():
    """
    Calcula o somatório dos valores de entrada.
    """

    def transforma(self, entradas):
        """
        Calcula o somatório dos valores de entrada.

        Args:
            entradas   : tupla com os valores a serem somados
        Ret:
            somatório dos valores de entrada
        """

        soma = 0
        for item in entradas:
            soma += item
        return soma

class Media():
    """
    Calcula a média dos valores de entrada.
    """

    def transforma(self, entradas):
        """
        Calcula a média dos valores de entrada.

        Args:
            entradas   : tupla com os valores dos quais calcular a média
        Ret:
            média dos valores de entrada
        """

        soma = f_somatorio.transforma(entradas)
        return soma/len(entradas)

class Agora():
    """
    Calcula o dia ou dia/hora atual do banco de dados ou do sistema 
    operacional.
    """

    def __init__(self, conexao=None, so_data=False):
        """
        Args:
            conexao : conexão com o banco de dados (caso se busque a
                      informação no banco) ou None se for para buscar no
                      sistema operacional
            so_data : flag se é para trazer só a data ou a hora também
        """
        self.__so_data = so_data
        self.__conn = conexao

    def transforma(self, entradas):
        """
        Calcula o dia/hora atual.

        Args:
            entradas   : não é necessário, mantido apenas para consistência da 
                         interface
        Ret:
            Dia (ou dia/hora) atual do banco de dados ou do sistema
            operacional, a depender de como o objeto foi construído.
        """

        if self.__conn is None:
            # busca do sistema operacional
            if self.__so_data:
                ret = date.today()
            else:
                ret = datetime.now().replace(microsecond=0)
        else:
            # busca do banco de dados
            if self.__so_data:
                ret = self.__conn.get_agora().replace(hour=0, minute=0, second=0, microsecond=0)
            else:
                ret = self.__conn.get_agora().replace(microsecond=0)

        return ret

class HTMLParaTxt():
    """
    Converte uma string com formatação em HTML (tags e acentuação) para um
    texto puro. As tags encontradas serão simplesmente eliminadas do texto
    final. 

    Como textos em HMTL tendem a ser grandes, opcionalmente a string de retorno
    pode ser truncada em uma determinada quantidade de bytes; este trunc leva
    em consideração os bytes necessários para acentuação em UTF-8.
    """
 
    def __init__(self, qtd_bytes=None):
        """
        Args:
        :param qtd_bytes: opcional; quantidade de bytes para um trunc da string
        final.
        """
        self.__qtd_bytes = qtd_bytes
 
    def transforma(self, entradas):
        """
        Retorna a string da entrada em HTML transformada para texto puro, com
        ou sem trunc. Se entrada for None, retorna None.

        # TODO: 
        se o HTML tiver os caracteres '<' ou '>' na sua parte textual essa
        função provavelmente falhará

        Args:
            entradas : tupla contendo a string a ser transformada
        """

        if len(entradas) != 1:
            raise RuntimeError(
                "Não pode haver mais de um dado de entrada.")

        if entradas[0] is None:
            return None

        ret = entradas[0]
        ret = ret.replace("\r", " ")
        ret = ret.replace("\n", " ")
        ret = ret.replace("\t", " ")

        # trata o html, corrigindo os acentos e retirando as tags
        ret = re.sub("(?<=<).*?(?=>)", "", html.unescape(ret))
        ret = ret.replace("\xa0", " ")
        ret = ret.replace("<>", "")

        # troca aspas simples do texto por duas aspas simples, para não dar
        # problema na hora da inserção no oracle
        ret = DeParaChar({"'": "''"}).transforma((ret, ))

        # retira excessos de espaços em branco
        ret = Trim().transforma((ret, ))
        ret = re.sub("\s{2,}", " ", ret)

        if self.__qtd_bytes is not None:
            ret = TruncaStringByte(self.__qtd_bytes).transforma((ret, ))

        return ret

class Trim():
    """
    Faz um trim numa string. Pode-se parametrizar se será um trim no início e
    fim, só no início ou só no fim.
    """
 
    def __init__(self, tipo_trim = "inicio_fim"):
        """
        Args:
        :param tipo_trim: tipo de trim a ser feito. Valores possíveis:
        "inicio_fim", "inicio", "fim"
        """
        self.__tipo_trim = tipo_trim
 
    def transforma(self, entradas):
        """
        Retorna a string da entrada com trim. Se entrada for None, retorna
        None.

        Args:
            entradas : tupla contendo a string a ser transformada
        """

        if len(entradas) != 1:
            raise RuntimeError(
                "Não pode haver mais de um dado de entrada para trim.")

        if entradas[0] is None:
            return None

        if self.__tipo_trim == "inicio_fim":
            trim = entradas[0].strip()
        elif self.__tipo_trim == "inicio":
            trim = entradas[0].ltrip()
        elif self.__tipo_trim == "fim":
            trim = entradas[0].rtrip()
        else:
            raise RuntimeError(
                "Tipo de trim incorreto.")

        return trim

class TruncaStringByte():
    """
    Trunca uma string até o número de bytes informado. Caracteres acentuados
    são considerados, de forma que é garantido que a quantidade de bytes
    utilizados para a string nunca é maior que a quantidade máxima de bytes
    informado no construtor.
    """
 
    def __init__(self, qtd_bytes):
        """
        Args:
        :param qtd_bytes: quantidade de bytes que a string terá ao final
        """
        self.__qtd_bytes = qtd_bytes
 
    def transforma(self, entradas):
        """
        Retorna a string truncada. Se entrada for None, retorna None.

        Args:
            entradas : tupla contendo a string a ser transformada
        """

        if len(entradas) != 1:
            raise RuntimeError(
                "Não pode haver mais de um dado de entrada para truncar.")

        if entradas[0] is None:
            return None
        
        # solução obtida em https://stackoverflow.com/questions/13665001/python-truncating-international-string
        if len(entradas[0].encode('utf-8')) > self.__qtd_bytes:
            trunc = \
                entradas[0].encode('utf-8')[:self.__qtd_bytes].decode('utf-8',
                'ignore')
        else: 
            trunc = entradas[0]

        return trunc

class LookupViaTabela():
    """
    Transforma o dado a partir de uma tabela de lookup informada.
    """
 
    def __init__(self, 
            conexao, 
            tabela_lookup, 
            campo, 
            chave,
            filtro = ""):
        """
        Args:
        :param conexao: conexão com o banco de dados que contém a tabela de
        lookup
        :param tabela_lookup: a tabela de lookup
        :param campo: o campo a ser retornado da lookup
        :param chave: o campo na tabela de lookup que liga as duas tabelas
        :param filtro: opcional; um filtro que pode ser aplicado ao montar a
        cláusula WHERE do SQL de busca

        Por exemplo, para buscar o texto de informação na tabela de informação
        da solução para o tipo 31 de uma determinada solução, ou seja, para
        executar o comando SQL abaixo:

        'select TX_INFORMACAO 
        from INFO_SOLUCAO 
        where
        ID_TIPO_INFORMACAO = 31 and ID_SOLUCAO = 2207'

        e considerando que a chave que liga as duas tabelas na lookup é
        ID_SOLUCAO, os parâmetros devem ser:

        tabela_lookup = 'INFO_SOLUCAO'
        campo = 'TX_INFORMACAO'
        chave = 'ID_SOLUCAO'
        filtro = 'ID_TIPO_INFORMACAO = 31'

        O valor de ID_SOLUCAO no SQL final será preenchido automaticamente a
        cada linha no loop de carga da tabela de origem.
        """

        self.__conexao = conexao
        self.__tabela_lookup = tabela_lookup
        self.__campo = campo
        self.__chave = chave
        self.__filtro = filtro
 
    def transforma(self, entradas):
        """
        Retorna o dado buscado numa tabela de lookup, usando a entrada
        informada como parâmetro de busca na lookup.

        Args:
            entradas : valor da chave de busca da lookup
        """

        # não poderia mesmo haver mais de uma entrada?
        if len(entradas) != 1:
            raise RuntimeError(
                    "Não pode haver mais de um dado de entrada para buscar na "
                    "tabela de lookup.")

        if entradas[0] is None:
            return None
        
        where = self.__filtro + " AND " \
            if self.__filtro != "" else self.__filtro
        sql =   "select " + self.__campo +              \
                " from " + self.__tabela_lookup +       \
                " where " + where +                     \
                self.__chave + " = " + str(entradas[0])

        try:
            registro = self.__conexao.consulta(sql)
        except:
            raise RuntimeError("Erro na execução do SELECT no banco de dados.")

        try:
            ret = next(registro)[0]
        except StopIteration:
            ret = None

        return ret



# TODO: implementar classe de transformação de conversão tipo

#  objetos de classes de transformação que só fazem sentido ter uma instância 
#  (singletons). Desta forma, ao importar este módulo estes objetos já estarão
#  instanciandos e bastará utilizá-los, sem necessidade de criar uma nova
#  instância no código que importou este módulo
f_copia     = Copia()
f_somatorio = Somatorio()
f_media     = Media()

